-- ***************************************************************************
-- ***************************************************************************
-- Copyright 2011(c) Analog Devices, Inc.
-- 
-- All rights reserved.
-- 
-- Redistribution and use in source and binary forms, with or without modification,
-- are permitted provided that the following conditions are met:
--     - Redistributions of source code must retain the above copyright
--       notice, this list of conditions and the following disclaimer.
--     - Redistributions in binary form must reproduce the above copyright
--       notice, this list of conditions and the following disclaimer in
--       the documentation and/or other materials provided with the
--       distribution.
--     - Neither the name of Analog Devices, Inc. nor the names of its
--       contributors may be used to endorse or promote products derived
--       from this software without specific prior written permission.
--     - The use of this software may or may not infringe the patent rights
--       of one or more patent holders.  This license does not release you
--       from the requirement that you obtain separate licenses from these
--       patent holders to use this software.
--     - Use of the software either in source or binary form, must be run
--       on or directly connected to an Analog Devices Inc. component.
--    
-- THIS SOFTWARE IS PROVIDED BY ANALOG DEVICES "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
-- INCLUDING, BUT NOT LIMITED TO, NON-INFRINGEMENT, MERCHANTABILITY AND FITNESS FOR A
-- PARTICULAR PURPOSE ARE DISCLAIMED.
--
-- IN NO EVENT SHALL ANALOG DEVICES BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
-- EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, INTELLECTUAL PROPERTY
-- RIGHTS, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR 
-- BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
-- STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF 
-- THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
-- ***************************************************************************
-- ***************************************************************************
-- This is a customized spi to cpld bridge (used for the 4dps boards)

library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_arith.all;
use ieee.std_logic_unsigned.all;

library unisim;
use unisim.all;

entity util_cpld is
  port (
    clk               : in  std_logic;

    -- 4-wire spi interface, usually for the xilinx core

    spi4_csn          : in  std_logic;
    spi4_clk          : in  std_logic;
    spi4_mosi         : in  std_logic;
    spi4_miso         : out std_logic;

    -- 3-wire spi interface, usually for the cpld core

    spi3_csn          : out std_logic;
    spi3_clk          : out std_logic;
    spi3_sdio_T       : out std_logic;
    spi3_sdio_O       : out std_logic;
    spi3_sdio_I       : in  std_logic;

    -- debug signals (chipscope)

    spi_dbg_trigger   : out std_logic;
    spi_dbg_data      : out std_logic_vector(28 downto 0)
  );
end util_cpld;

architecture IMP of util_cpld is

  -- predefined spi mux for the different devices on board

  signal FMC27X_CPLD      : std_logic_vector(7 downto 0) := "00000000";
  signal FMC27X_AD9517    : std_logic_vector(7 downto 0) := "10000100";
  signal FMC27X_AD9250_0  : std_logic_vector(7 downto 0) := "10000000";
  signal FMC27X_AD9250_1  : std_logic_vector(7 downto 0) := "10000001";
  signal FMC27X_AD9129_0  : std_logic_vector(7 downto 0) := "10000010";
  signal FMC27X_AD9129_1  : std_logic_vector(7 downto 0) := "10000011";

  signal spi4_clk_d       : std_logic := '0';
  signal spi4_csn_d       : std_logic := '0';
  signal spi4_clkcnt      : std_logic_vector(5 downto 0) := (others => '0');
  signal spi4_bitcnt      : std_logic_vector(6 downto 0) := (others => '0');
  signal spi4_devid       : std_logic_vector(7 downto 0) := (others => '0');
  signal spi4_rwn         : std_logic := '0';
  signal spi3_enable      : std_logic := '0';


begin

  FMC27X_CPLD <= "00000000";
  FMC27X_AD9517 <= "10000100";
  FMC27X_AD9250_0 <= "10000000";
  FMC27X_AD9250_1 <= "10000001";
  FMC27X_AD9129_0 <= "10000010";
  FMC27X_AD9129_1 <= "10000011";

  -- pass through most of the stuff (no need to change clock or miso or mosi)

  spi4_miso <= spi3_sdio_I;
  spi3_csn <= spi4_csn;
  spi3_clk <= spi4_clk;
  spi3_sdio_T <= not(spi4_csn) and spi3_enable;
  spi3_sdio_O <= spi4_mosi;

  -- debug signals (chipscope)

  spi_dbg_trigger <= spi4_csn;
  spi_dbg_data(28) <= spi4_clk_d;
  spi_dbg_data(27) <= spi4_csn_d;
  spi_dbg_data(26 downto 21) <= spi4_clkcnt;
  spi_dbg_data(20 downto 14) <= spi4_bitcnt;
  spi_dbg_data(13 downto 6) <= spi4_devid;
  spi_dbg_data(5) <= spi4_rwn;
  spi_dbg_data(4) <= spi3_enable;
  spi_dbg_data(3) <= spi4_csn;
  spi_dbg_data(2) <= spi4_clk;
  spi_dbg_data(1) <= spi4_mosi;
  spi_dbg_data(0) <= spi3_sdio_I;

  -- the spi4 format is a preamble that selects a particular device, so all we need
  -- to do is collect the first 8 bits, then control the tristate based on the
  -- device's address and data widths. the details of the spi formats can be found
  -- in the data sheet of the devices.

  p_cnts: process(clk) begin
    if (clk'event and clk = '1') then
      spi4_clk_d <= spi4_clk;
      spi4_csn_d <= spi4_csn;
      if ((spi4_clk = '1') and (spi4_clk_d = '0')) then
        spi4_clkcnt <= (others => '0');
      else
        spi4_clkcnt <= spi4_clkcnt + 1;
      end if;
      if ((spi4_csn = '1') and (spi4_csn_d = '0')) then
        spi4_bitcnt <= (others => '0');
        spi4_devid <= (others => '0');
        spi4_rwn <= '0';
      elsif ((spi4_clk = '1') and (spi4_clk_d = '0')) then
        spi4_bitcnt <= spi4_bitcnt + 1;
        if (spi4_bitcnt < 8) then
          spi4_devid <= spi4_devid(6 downto 0) & spi4_mosi;
        end if;
        if (spi4_bitcnt = 8) then
          spi4_rwn <= spi4_mosi;
        end if;
      end if;
      if (spi4_csn = '0') then
        if ((spi4_devid = FMC27X_CPLD) or (spi4_devid = FMC27X_AD9129_0) or
          (spi4_devid = FMC27X_AD9129_1)) then
          if ((spi4_bitcnt = 16) and (spi4_clkcnt = 8)) then
            spi3_enable <= spi4_rwn;
          end if;
        elsif ((spi4_devid = FMC27X_AD9517) or (spi4_devid = FMC27X_AD9250_0) or 
          (spi4_devid = FMC27X_AD9250_1)) then
          if ((spi4_bitcnt = 24) and (spi4_clkcnt = 8)) then
            spi3_enable <= spi4_rwn;
          end if;
        end if;
      else
        spi3_enable <= '0';
      end if;
    end if;
  end process;

end IMP;

-- ***************************************************************************
-- ***************************************************************************
